"""
Copyright (c) 2024 OTB Africa. All rights reserved. <https://otbafrica.com>

Created Date: Sunday, May 5th 2024, 8:00:41 pm
Author: OTB Africa, developers@otbafrica.com

Use of this source code is governed by the license that can be found in 
the LICENSE file or at https://developers.otbafrica.com/licenses/
"""

from django.db import models
from django.utils.translation import gettext_lazy as _

from django_fsm import transition, FSMField, can_proceed

from apps.digital_signatures.models import DigitalSignatureFolder


class OpenSignTrackerQuerySet(models.QuerySet):
    class STATUS(models.TextChoices):
        PENDING = 'PENDING', _('Pending')
        SIGNED = 'SIGNED', _('Signed')
        CANCELLED = 'CANCELLED', _('Cancelled')

    def signed(self):
        return self.filter(status=self.STATUS.SIGNED)
    
    def pending(self):
        return self.filter(status=self.STATUS.PENDING)
    
    def cancelled(self):
        return self.filter(status=self.STATUS.CANCELLED)

class OpenSignDocumentTracker(models.Model):
    """
    Model for tracking OpenSign document. Since OpenSign does not support signing multiple
    documents at once, we cannot effectively use the digital signing folder to sync the status
    between opensign and onestop.
    """

    generated_document = models.ForeignKey('onestopdocuments.GeneratedDocument', on_delete=models.CASCADE)
    opensign_document_id = models.CharField(max_length=255)
    status = FSMField(max_length=20, choices=OpenSignTrackerQuerySet.STATUS.choices, default=OpenSignTrackerQuerySet.STATUS.PENDING)

    objects = OpenSignTrackerQuerySet.as_manager()

    class Meta:
        db_table = 'opensign_document_tracker'
        unique_together = ('generated_document', 'opensign_document_id', 'status')
    
    @transition(field=status, source=OpenSignTrackerQuerySet.STATUS.PENDING, target=OpenSignTrackerQuerySet.STATUS.SIGNED)
    def mark_as_signed(self):
        self.status = OpenSignTrackerQuerySet.STATUS.SIGNED
    
    def _cancellation_check(self) -> bool:
        digital_signature_folder: DigitalSignatureFolder = self.generated_document.digital_signature_folder
        folder_is_good_to_go = digital_signature_folder and digital_signature_folder.status in [
            digital_signature_folder.SIGNING_FOLDER_STATUS_CHOICES.pending,
            digital_signature_folder.SIGNING_FOLDER_STATUS_CHOICES.processing,
        ]
        no_one_has_signed = not self.signer_trackers.signed().exists()
        return folder_is_good_to_go and no_one_has_signed

    @transition(field=status, source=OpenSignTrackerQuerySet.STATUS.PENDING, target=OpenSignTrackerQuerySet.STATUS.CANCELLED, conditions=[_cancellation_check])
    def cancel(self):
        self.status = OpenSignTrackerQuerySet.STATUS.CANCELLED
        for signing_tracker in self.signer_trackers.all():
            signing_tracker.cancel()
            signing_tracker.save()

    def can_cancel(self) -> bool:
        return can_proceed(self.cancel)

    @property
    def has_signed(self):
        return self.status == OpenSignTrackerQuerySet.STATUS.SIGNED


class OpenSignSignerTracker(models.Model):
    
    document_tracker = models.ForeignKey(OpenSignDocumentTracker, on_delete=models.CASCADE, related_name='signer_trackers')
    signer_email = models.EmailField()
    signer_url = models.URLField()
    status = FSMField(max_length=20, choices=OpenSignTrackerQuerySet.STATUS.choices, default=OpenSignTrackerQuerySet.STATUS.PENDING)

    objects = OpenSignTrackerQuerySet.as_manager()

    class Meta:
        db_table = 'opensign_signer_tracker'
        unique_together = ('document_tracker', 'signer_email')

    @transition(field=status, source=OpenSignTrackerQuerySet.STATUS.PENDING, target=OpenSignTrackerQuerySet.STATUS.SIGNED)
    def mark_as_signed(self):
        self.status = OpenSignTrackerQuerySet.STATUS.SIGNED
    
    @transition(field=status, source=OpenSignTrackerQuerySet.STATUS.PENDING, target=OpenSignTrackerQuerySet.STATUS.CANCELLED)
    def cancel(self):
        self.status = OpenSignTrackerQuerySet.STATUS.CANCELLED



"""
Copyright (c) 2024 OTB Africa. All rights reserved. <https://otbafrica.com>

Created Date: Tuesday, April 30th 2024, 9:04:46 pm
Author: OTB Africa, developers@otbafrica.com

Use of this source code is governed by the license that can be found in 
the LICENSE file or at https://developers.otbafrica.com/licenses/
"""

from config.settings.environment import env

OPENSIGN_API_BASE_URL = env.str('OPENSIGN_API_BASE_URL', 'http://opensign.localhost/v1/')
OPENSIGN_API_TOKEN = env.str('OPENSIGN_API_TOKEN', None)




"""
Copyright (c) 2024 OTB Africa. All rights reserved. <https://otbafrica.com>

Created Date: Tuesday, April 30th 2024, 8:26:14 pm
Author: OTB Africa, developers@otbafrica.com

Use of this source code is governed by the license that can be found in 
the LICENSE file or at https://developers.otbafrica.com/licenses/
"""

INSTALLED_APPS = [
    'onestop_opensign',
]

OS_DIGITAL_SIGNATURE_EXTENSIONS = ({
    'opensign': {
        'DIGITAL_SIGNATURE_PROVIDER_NAME': "OpenSign",
        'DIGITAL_SIGNATURE_PROVIDER_CLASS': "onestop_opensign.provider.OpenSignSignatureProvider"
    },
})



"""
Copyright (c) 2024 OTB Africa. All rights reserved. <https://otbafrica.com>

Created Date: Thursday, May 2nd 2024, 7:40:29 am
Author: OTB Africa, developers@otbafrica.com

Use of this source code is governed by the license that can be found in 
the LICENSE file or at https://developers.otbafrica.com/licenses/
"""

from functools import reduce
import logging
import requests

from django.utils.decorators import method_decorator
from rest_framework.response import Response
from rest_framework.views import APIView

from apps.base.middleware.login import login_exempt
from apps.digital_signatures.models import DigitalSignatureFolder
from apps.onestopauth.users import get_user_model
from apps.onestopdocuments import utils as document_utils
from .models import OpenSignDocumentTracker

UserModel = get_user_model()


_logger = logging.getLogger(__name__)

@method_decorator(login_exempt, name='dispatch')
class OpenSignSignatureCallbackView(APIView):

    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except Exception as e:
            _logger.error(f"Error occurred in OpenSignSignatureCallbackView: {e}")
            return Response(dict(status='error', message='An error occurred'))

    def post(self, request, *args, **kwargs):
        data = request.data
        try:
            document_tracker = OpenSignDocumentTracker.objects.get(opensign_document_id=data.get('objectId'))
            onestop_document = document_tracker.generated_document
            digital_signature_folder: DigitalSignatureFolder = onestop_document.digital_signature_folder
        except OpenSignDocumentTracker.DoesNotExist:
            _logger.error(f"Document not found for OpenSign ID: {data.get('objectId')}")
            return Response(dict(status='error', message='Document not found'))

        if data.get('event') == 'signed' and (opensign_signer := data.get('signer')):
            signer_email = opensign_signer.get('email')
            signing_user = UserModel.objects.get(email=signer_email)

            if document_tracker:
                opensign_signer_tracker = document_tracker.signer_trackers.get(signer_email=signer_email)
                opensign_signer_tracker.mark_as_signed()
                opensign_signer_tracker.save()

            digital_signature_folder.start_signing_session()
            digital_signature_folder.save()

            folder_documents = digital_signature_folder.generateddocument_set.all()
            folder_document_trackers = OpenSignDocumentTracker.objects.filter(generated_document__in=folder_documents)
            pending_signer_trackers = reduce(lambda prev, curr: prev | curr, [document_tracker.signer_trackers.pending().filter(signer_email=signer_email) for document_tracker in folder_document_trackers])
            if not pending_signer_trackers.exists():
                digital_signature_folder.record_user_signature(signing_user)
                digital_signature_folder.save()

            _logger.info(f"Signer {opensign_signer} signed successfully")

        elif data.get('event') == 'completed':
            try:
                signed_file_http_response = requests.get(data.get('file'))
                document_utils.save_signed_document_from_bytes(
                    document=onestop_document,
                    signed_file_content=signed_file_http_response.content,
                    save_directory='opensign/'
                )

                if not digital_signature_folder.has_documents_pending_signature():
                    digital_signature_folder.complete_signing_session()    
                    digital_signature_folder.save()
                    document_tracker.mark_as_signed()
                    document_tracker.save()
                _logger.info(f"Document {onestop_document.document_number} signed successfully")
            except Exception as e:
                _logger.error(f"Error occurred while processing signed document: {e}")
        return Response(dict(status='ok'))




"""
Copyright (c) 2024 OTB Africa. All rights reserved. <https://otbafrica.com>

Created Date: Sunday, May 5th 2024, 8:00:52 pm
Author: OTB Africa, developers@otbafrica.com

Use of this source code is governed by the license that can be found in 
the LICENSE file or at https://developers.otbafrica.com/licenses/
"""





# Generated by Django 4.2.2 on 2024-05-08 08:18

from django.db import migrations, models
import django.db.models.deletion
import django_fsm


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('onestopdocuments', '0003_auto_20221002_1251'),
    ]

    operations = [
        migrations.CreateModel(
            name='OpenSignDocumentTracker',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('opensign_document_id', models.CharField(max_length=255)),
                ('status', django_fsm.FSMField(choices=[('PENDING', 'Pending'), ('SIGNED', 'Signed'), ('CANCELLED', 'Cancelled')], default='PENDING', max_length=20)),
                ('generated_document', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='onestopdocuments.generateddocument')),
            ],
            options={
                'db_table': 'opensign_document_tracker',
                'unique_together': {('generated_document', 'opensign_document_id', 'status')},
            },
        ),
        migrations.CreateModel(
            name='OpenSignSignerTracker',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('signer_email', models.EmailField(max_length=254)),
                ('signer_url', models.URLField()),
                ('status', django_fsm.FSMField(choices=[('PENDING', 'Pending'), ('SIGNED', 'Signed'), ('CANCELLED', 'Cancelled')], default='PENDING', max_length=20)),
                ('document_tracker', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='signer_trackers', to='onestop_opensign.opensigndocumenttracker')),
            ],
            options={
                'db_table': 'opensign_signer_tracker',
                'unique_together': {('document_tracker', 'signer_email')},
            },
        ),
    ]



"""
Copyright (c) 2024 OTB Africa. All rights reserved. <https://otbafrica.com>

Created Date: Tuesday, April 30th 2024, 8:28:27 pm
Author: OTB Africa, developers@otbafrica.com

Use of this source code is governed by the license that can be found in 
the LICENSE file or at https://developers.otbafrica.com/licenses/
"""

import json
import logging
import requests
from typing import List, TypedDict, Optional

from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.http import JsonResponse

from apps.base import utils as base_utils
from apps.digital_signatures.provider import AbstractSignatureProvider
from apps.onestopauth.models import Team
from apps.onestopdocuments import utils as document_utils

from . import config
from .models import OpenSignDocumentTracker

_logger = logging.getLogger(__name__)


class OpenSignSignerWidgetOptions(TypedDict):
    required: bool
    name: str


class OpenSignSignerWidget(TypedDict):
    type: str
    page: int
    x: int
    y: int
    w: int
    h: int
    options: Optional[OpenSignSignerWidgetOptions]


class OpenSignDocumentSigner(TypedDict):
    name: str
    email: str
    phone: str
    widgets: List[OpenSignSignerWidget]


class OpenSignDocumentSignURL(TypedDict):
    email: str
    url: str


class OpenSignCreateDocumentPayload(TypedDict):
    file: str
    title: str
    description: str
    note: str
    signers: List[OpenSignDocumentSigner]


class OpenSignCreateDraftDocumentResponse(TypedDict):
    objectId: str
    url: str


class OpenSignCreateDocumentResponse(TypedDict):
    objectId: str
    message: str
    signurl: List[OpenSignDocumentSignURL]


class OpenSignAPIClient:
    api_token = config.OPENSIGN_API_TOKEN

    @property
    def base_url(self):
        base_url = config.OPENSIGN_API_BASE_URL
        return base_url if base_url.endswith('/') else f"{base_url}/"
    
    def get_endpoint(self, resource: str):
        return f"{self.base_url}{resource}"
    
    def get_create_document_data(self, base_64_document: str, title: str, signers: List[OpenSignDocumentSigner], description, note) -> OpenSignCreateDocumentPayload:
        return dict(
            file=base_64_document.decode(),
            title=title,
            description=description,
            note=note,
            signers=signers
        )

    def create_draft_document(self, base_64_document: str, title: str, signers: List[OpenSignDocumentSigner], **kwargs) -> OpenSignCreateDraftDocumentResponse:
        description = kwargs.get('description', '')
        note = kwargs.get('note', '')
        data=json.dumps(self.get_create_document_data(
            base_64_document=base_64_document,
            title=title,
            signers=signers,
            description=description,
            note=note
        ))
        response = requests.post(
            self.get_endpoint('draftdocument'),
            headers={
                'x-api-token': self.api_token,
                'Content-Type': 'application/json',
                'Accept': 'application/json'
            },
            data=data
        )
        _logger.info(f"OpenSign: Response for Create Draft Document: {response.json()}\n\n")
        return response.json()

    def create_document(self, base_64_document: str, title: str, signers: List[OpenSignDocumentSigner], **kwargs) -> OpenSignCreateDocumentResponse:
        description = kwargs.get('description', '')
        note = kwargs.get('note', '')
        data=json.dumps(self.get_create_document_data(
            base_64_document=base_64_document,
            title=title,
            signers=signers,
            description=description,
            note=note
        ))

        response = requests.post(
            self.get_endpoint('createdocument'),
            headers={
                'x-api-token': self.api_token,
                'Content-Type': 'application/json',
                'Accept': 'application/json'
            },
            data=data
        )
        _logger.info(f"OpenSign: Response for Create Document: {response.json()}\n\n")
        return response.json()
    
    def delete_document(self, opensign_document_id: str):
        response = requests.delete(
            self.get_endpoint(f'document/{opensign_document_id}'),
            headers={
                'x-api-token': self.api_token,
                'Content-Type': 'application/json',
                'Accept': 'application/json'
            }
        )
        _logger.info(f"OpenSign: Response for Delete Document: {response.json()}\n\n")
        return response.json()


class OpenSignSignatureProvider(AbstractSignatureProvider):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.api_client = OpenSignAPIClient()
    
    def create_tracker(self, document, opensign_document_id, signer_urls: List[OpenSignDocumentSignURL]) -> OpenSignDocumentTracker:
        tracker = OpenSignDocumentTracker.objects.create(
            generated_document=document,
            opensign_document_id=opensign_document_id
        )
        for signer_url in signer_urls:
            tracker.signer_trackers.create(
                signer_email=signer_url.get('email'),
                signer_url=signer_url.get('url')
            )
        return tracker

    def get_tracker(self, document) -> OpenSignDocumentTracker:
        try:
            return OpenSignDocumentTracker.objects.pending().get(generated_document=document)
        except OpenSignDocumentTracker.DoesNotExist:
            pass
        return None

    def get_signer_details(self, user, digital_signature_folder) -> OpenSignDocumentSigner:
        widgets = self.get_widgets_from_settings(user, digital_signature_folder)
        try:
            registration_profile = user.registrationprofile
            return {
                'name': user.profile_name,
                'email': user.email,
                'phone': str(registration_profile.phone_number),
                "widgets": widgets
            }
        except ObjectDoesNotExist:
            pass
        return None
    
    def get_widgets_from_settings(self, user, digital_signature_folder) -> list:
        """
        Get widgets set in the opensign settings for the task template in the workflow.
        OpenSign requires widgets to be set before a user can sign. As per documentation, the
        format for widgets is as follows:
                "widgets": [
                    {
                    "type": "signature",
                    "page": 1,
                    "x": 244,
                    "y": 71,
                    "w": 38,
                    "h": 46
                    }
                ]
        https://docs.opensignlabs.com/docs/API-docs/createdocument
        """
        widgets = []
        try:
            digital_signing_settings = digital_signature_folder.signing_task_template.digital_signing_settings
            open_sign_settings = digital_signing_settings.settings.get('open_sign', {})

            for key, value in open_sign_settings.items():
                try:
                    team = base_utils.get_object_or_none(Team, pk=key)
                    _logger.info(f"Checking if signer is in team: {team}, set for signing...")
                except Exception as e:
                    _logger.error(f"Error occurred while getting team: {e}")
                    team = None

                if user.email == key or (team and user in team.members):
                    widgets.extend(value.get('widgets', []))

        except Exception as e:
            _logger.error(f"Error occurred while getting digital signature settings: {e}")

        return widgets

    def sign_documents(self, digital_signature_folder, request, **kwargs):
        """
        Method that signs document(s) immediately with out sending to a folder
        """
        signers = list(map(lambda user: self.get_signer_details(user, digital_signature_folder), digital_signature_folder.signers.all()))
        sign_urls = []

        for document in digital_signature_folder.generateddocument_set.all():
            doc_as_b64_string, content_type, extension, file_name = document_utils.get_pdf_as_string(
                document=document,
                cnt_type='base64',
                user=request.user
            )

            tracker = self.get_tracker(document)
            if tracker and tracker.can_cancel():
                self.api_client.delete_document(tracker.opensign_document_id)
                tracker.cancel()
                tracker.save()
                tracker = None
            elif tracker:
                opensign_document_id = tracker.opensign_document_id
                sign_urls.extend(list(map(lambda signer_tracker: dict(email=signer_tracker.signer_email, url=signer_tracker.signer_url), tracker.signer_trackers.pending())))


            if tracker is None:
                response_body = self.api_client.create_document(
                    base_64_document=doc_as_b64_string,
                    title=file_name,
                    signers=signers
                )
                
                opensign_document_id = response_body.get('objectId')
                tracker = self.create_tracker(
                    document=document,
                    opensign_document_id=opensign_document_id,
                    signer_urls=response_body.get('signurl', [])
                )
                sign_urls.extend(response_body.get('signurl', []))

        filtered_sign_urls = list(filter(lambda sign_url: sign_url.get('email') == request.user.email, sign_urls))
        if filtered_sign_urls:
            sign_url = filtered_sign_urls[0].get('url')
            return JsonResponse(dict(redirect_url=sign_url))
        
        process_instance = kwargs.get('context', {}).get('process')
        messages.info(request, f"Document sent for signing with OpenSign. You are not among any of the signers.")
        return JsonResponse(dict(redirect_url=process_instance.get_absolute_url() if process_instance else ''))

    def sign_folder_documents(self, digital_signature_folder, request, **kwargs):
        return self.sign_documents(digital_signature_folder, request, **kwargs)

    def on_user_add_to_folder(self, *args, **kwargs):
        """
        Do nothing since OpenSign does not support multiple docs per signing
        """
        pass

    def on_auto_add_to_folder(self, *args, **kwargs):
        """
        Do nothing since OpenSign does not support multiple docs per signing
        """
        pass



from setuptools import setup

"""
Fixtures folder has been included in Manifest.in file.
https://stackoverflow.com/questions/1612733/including-non-python-files-with-setup-py
https://docs.python.org/2/distutils/sourcedist.html#principle
"""

setup(
    name='onestop_opensign',
    version='0.1.5',
    author='OTB Africa',
    author_email='developers@otbafrica.com',
    license='BSD 2-clause',
    description='Package for a onestop extension that integrates with the opensign digital signature service.',
    packages=[
        'onestop_opensign',
        'onestop_opensign.migrations'
    ],
    include_package_data=True, # Include files from MANIFEST.in e.g. fixtures
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'requests',
        'django',
        'django-fsm'
    ],
)


"""
Copyright (c) 2024 OTB Africa. All rights reserved. <https://otbafrica.com>

Created Date: Thursday, May 2nd 2024, 7:40:56 am
Author: OTB Africa, developers@otbafrica.com

Use of this source code is governed by the license that can be found in 
the LICENSE file or at https://developers.otbafrica.com/licenses/
"""

from django.urls import path

from .views import OpenSignSignatureCallbackView


urlpatterns = [
    path('TkuZhGCITH/', OpenSignSignatureCallbackView.as_view(), name='opensign-signature-callback')
]
